#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "assert.h"

#include "List.h"

#include <FL/Fl.H>
#include <FL/Fl_Shared_Image.H>

#include "GMapTile.h"
#include "GMap.h"


extern int n_objects;
extern char *object_name[];
extern int object_parameters[];
extern char *object_parameter_name[36][3];
extern int object_parameters_default[36][3];
extern List<GMapTile> tiles;

GMapLayer::GMapLayer()
{
} /*  GMapLayer::GMapLayer */ 


GMapLayer::GMapLayer(FILE *fp,List<GMapTile> *tiles)
{
	int i,j;
	int nt,no,np,i_tmp,x,y;
	char tmp[256];
	GMapTilePlace *tp;
	GObjectPlace *op;
	fscanf(fp,"%i",&nt);

	for(i=0;i<nt;i++) {
		tp=new GMapTilePlace();
		fscanf(fp,"%i %i %s",&x,&y,tmp);
		tp->m_x=x;
		tp->m_y=y;
		tp->m_tile=0;

		{
			GMapTile *t;
			tiles->Rewind();
			while(tp->m_tile==0 && tiles->Iterate(t)) if (strcmp(t->get_name(),tmp)==0) tp->m_tile=t;
			assert(tp->m_tile!=0);
		}
		m_tiles.Add(tp);
	} // for 

	fscanf(fp,"%i",&no);
	for(i=0;i<no;i++) {
		op=new GObjectPlace();
		fscanf(fp,"%i %i %s %i",&x,&y,tmp,&np);
		op->m_x=x;
		op->m_y=y;
		op->m_name=new char[strlen(tmp)+1];
		strcpy(op->m_name,tmp);
		for(j=0;j<np;j++) {
			fscanf(fp,"%i",&i_tmp);
			op->m_parameters.Add(new int(i_tmp));
		} // for 
		m_objects.Add(op);
	} // for 
} /* GMapLayer::GMapLayer */ 


GMapLayer::~GMapLayer()
{
} /*  GMapLayer::~GMapLayer */ 


void GMapLayer::save(FILE *fp)
{
	int i;
	GMapTilePlace *tp;
	GObjectPlace *op;

	fprintf(fp,"%i\n",m_tiles.Length());

	m_tiles.Rewind();
	while(m_tiles.Iterate(tp)) {
		fprintf(fp,"%i %i %s\n",tp->m_x,tp->m_y,tp->m_tile->get_name());
	} // while 

	fprintf(fp,"%i\n",m_objects.Length());
	m_objects.Rewind();
	while(m_objects.Iterate(op)) {
		fprintf(fp,"%i %i %s %i",op->m_x,op->m_y,op->m_name,op->m_parameters.Length());
		for(i=0;i<op->m_parameters.Length();i++) fprintf(fp," %i",*(op->m_parameters[i]));
		fprintf(fp,"\n");
	} // while 
} /* GMapLayer::save */ 


void GMapLayer::draw(int dx,int dy)
{
	GMapTilePlace *tp;
	GObjectPlace *op;

	m_tiles.Rewind();
	while(m_tiles.Iterate(tp)) {
		tp->m_tile->get_sfc()->draw(tp->m_x+dx,tp->m_y+dy);
	} /* while */ 

	m_objects.Rewind();
	while(m_objects.Iterate(op)) {
		GMap::object_tile(op->m_name,&(op->m_parameters))->get_sfc()->draw(op->m_x+dx,op->m_y+dy);
	} /* while */ 


} /* GMapLayer::draw */ 


void GMapLayer::clear(void)
{
	m_tiles.Delete();
	m_objects.Delete();
} /* GMapLayer::add_time */ 


void GMapLayer::add_tile(int x,int y,GMapTile *tile)
{
	GMapTilePlace *tp=new GMapTilePlace();
	tp->m_tile=tile;
	tp->m_x=x;
	tp->m_y=y;
	m_tiles.Add(tp);
} /* GMapLayer::add_time */ 


void GMapLayer::add_object(int x,int y,char *name)
{
	int n=-1;
	int i;

	for(i=0;i<n_objects;i++) if (strcmp(object_name[i],name)==0) n=i;

	if (n==-1) return;

	GObjectPlace *op=new GObjectPlace();
	op->m_name=new char[strlen(name)+1];
	strcpy(op->m_name,name);
	op->m_x=x;
	op->m_y=y;
	for(i=0;i<object_parameters[n];i++) {
		op->m_parameters.Add(new int(object_parameters_default[n][i]));
	} /* for */ 
	m_objects.Add(op);
} /* GMapLayer::add_object */ 


bool GMapLayer::delete_tile(int x,int y)
{
	GMapTilePlace *tp;

	m_tiles.Rewind();
	while(m_tiles.Iterate(tp)) {
		if (tp->m_x<=x && tp->m_y<=y &&
			tp->m_x+tp->m_tile->get_sfc()->w()>x && tp->m_y+tp->m_tile->get_sfc()->h()>y) {
			m_tiles.DeleteElement(tp);
			delete tp;
			return true;
		} /* if */ 
	} /* while */ 

	return false;
} /* GMapLayer::delete_tile */ 


void GMapLayer::delete_object(GObjectPlace *o)
{
	if (m_objects.MemberRefP(o)) {
		m_objects.DeleteElement(o);
		delete o;
	} // if
} /* GMapLayer::delete_object */ 


GObjectPlace *GMapLayer::get_object(int x,int y)
{
	GObjectPlace *op;
	GMapTile *t;


	m_objects.Rewind();
	while(m_objects.Iterate(op)) {
		t=GMap::object_tile(op->m_name,&(op->m_parameters));

		if (op->m_x<=x && op->m_y<=y &&
			op->m_x+t->get_sfc()->w()>x && op->m_y+t->get_sfc()->h()>y) {
			return op;
		} /* if */ 
	} /* while */ 

	return 0;
} /* GMapLayer::get_object */ 


GMap::GMap(int dx,int dy)
{
	m_dx=dx;
	m_dy=dy;

	m_layers.Add(new GMapLayer());
	m_layers.Add(new GMapLayer());
	m_layers.Add(new GMapLayer());
	m_layers.Add(new GMapLayer());
	m_layers.Add(new GMapLayer());

	m_scripts=0;
} /* GMap:GMap */ 


GMap::GMap(FILE *fp,List<GMapTile> *tiles)
{
	GMapLayer *l;
	int i;
	int nl;
	fscanf(fp,"%i %i %i",&m_dx,&m_dy,&nl);

	for(i=0;i<nl;i++) {
		l=new GMapLayer(fp,tiles);
		m_layers.Add(l);
	} // for 

	{		
		int len=64;
		char *tmp;

		i=0;
		tmp=new char[len];
		while(!feof(fp)) {
			tmp[i++]=fgetc(fp);
			if (i==1 && (tmp[0]=='\n' || tmp[0]=='\r')) i=0;
			if (i>=len) {
				char *tmp2=new char[len+64];
				int j;
				for(j=0;j<len;j++) tmp2[j]=tmp[j];
				len+=64;
				delete []tmp;
				tmp=tmp2;
			} // if 
		} // while 
		tmp[i-1]=0;

		m_scripts=new char[strlen(tmp)+1];
		strcpy(m_scripts,tmp);
		delete []tmp;
	}
} /* GMap::GMap */ 


GMap::~GMap()
{
} /* GMap::~GMap */ 


void GMap::save(FILE *fp)
{
	GMapLayer *l;

	fprintf(fp,"%i %i\n%i\n",m_dx,m_dy,m_layers.Length());
	m_layers.Rewind();
	while(m_layers.Iterate(l)) l->save(fp);

	if (m_scripts!=0) fprintf(fp,"%s\n",m_scripts);
} /* GMap::save */ 


void GMap::draw(int dx,int dy)
{
	GMapLayer *l;

	m_layers.Rewind();
	while(m_layers.Iterate(l)) l->draw(dx,dy);
} /* GMap::draw */ 


void GMap::draw(int dx,int dy,int layer)
{
	if (layer>=0 && layer<m_layers.Length()) m_layers[layer]->draw(dx,dy);
} /* GMap::draw */ 


void GMap::add_tile(int layer,int x,int y,GMapTile *tile)
{
	if (layer>=0 && layer<m_layers.Length()) {
		m_layers[layer]->add_tile(x,y,tile);
	} /* if */ 
} /* GMap::add_time */ 


void GMap::add_object(int layer,int x,int y,char *name)
{
	if (layer>=0 && layer<m_layers.Length()) {
		m_layers[layer]->add_object(x,y,name);
	} /* if */ 
} /* GMap::add_object */ 



void GMap::delete_layer(int layer)
{
	m_layers.DeletePosition(layer);
} /* GMap::delete_layer */ 


void GMap::clear_layer(int layer)
{
	m_layers[layer]->clear();
} /* GMap::delete_layer */ 



void GMap::add_layer(void)
{
	m_layers.Add(new GMapLayer());
} /* GMap::add_layer */ 



bool GMap::delete_tile(int x,int y)
{
	GMapLayer *l;

	m_layers.Rewind();
	while(m_layers.Iterate(l)) if (l->delete_tile(x,y)) return true;
	return false;
} /* GMap::delete_tile */ 



bool GMap::delete_tile(int layer,int x,int y)
{
	if (layer>=0 && layer<m_layers.Length()) return m_layers[layer]->delete_tile(x,y);

	return false;
} /* GMap::delete_tile */ 


void GMap::delete_object(GObjectPlace *o)
{
	GMapLayer *l;

	m_layers.Rewind();
	while(m_layers.Iterate(l)) l->delete_object(o);
} /* GMap::delete_object */ 


GObjectPlace *GMap::get_object(int x,int y)
{
	GObjectPlace *op;
	GMapLayer *l;

	m_layers.Rewind();
	while(m_layers.Iterate(l)) {
		op=l->get_object(x,y);
		if (op!=0) return op;
	} // while 
	return 0;
} /* GMap::get_object */ 


GObjectPlace *GMap::get_object(int layer,int x,int y)
{
	if (layer>=0 && layer<m_layers.Length()) return m_layers[layer]->get_object(x,y);

	return 0;
} /* GMap::get_object */ 


GMapTile *GMap::object_tile(char *name,List<int> *parameters)
{
	int i;
	int n=-1;
	char *tile[]={"ob_key","ob_falling-rock1","ob_skull-l1","ob_skull-l1",
				  "ob_cagedoor-closed","ob_drop3","ob_skull-entrance","ob_water-opening",
				  "ob_water-opening-blue","ob_water-opening-green",
				  "ob_entrydoor","ob_rope-top","ob_rope","ob_rope-bottom",
				  "ob_water","ob_falling-bluerock1","ob_falling-greenrock1","ob_rock",
				  "ob_bat1","ob_skeleton-l1","ob_falling-yellowrock1","ob_water-opening-yellow",
				  "ob_rock_blue","ob_rock_green","ob_rock_yellow",
				  "ob_pipe-water-l4","ob_pipe-water-r4","ob_skull-l1",
				  "ob_lava","ob_exitdoor","wl_tile-wall-l2","wl_tileblue-wall-l2","wl_tilegreen-wall-l2","wl_tileyellow-wall-l2",
				  "ob_flame2","ob_ghost1"};

	for(i=0;i<n_objects;i++) if (strcmp(object_name[i],name)==0) n=i;

	if (n==-1) return 0;

	GMapTile *t;
	tiles.Rewind();
	while(tiles.Iterate(t)) {
		if (strcmp(t->get_name(),tile[n])==0) {
			return t;
		} // if 
	} /* while */ 

	return 0;
} /* GMap::draw_object */ 


void GMap::set_scripts(char *text)
{
	if (m_scripts!=0) delete []m_scripts;

	if (text==0) {
		m_scripts=0;
	} else {
		m_scripts=new char[strlen(text)+1];
		strcpy(m_scripts,text);
	} // if 
} /* GMap::set_scripts */ 



